#!/usr/bin/perl

#
# $Id: lipm_Nstreches_to_gff3.pl 1613 2018-07-16 15:43:32Z gouzy $
#

=pod

=head1

=head1 NAME

=head1 MANDATORY

=head1 

=cut

BEGIN
{
    my ($dirprg, $nameprg) = $0 =~ /(.+)\/(.+)/;
	unshift @INC, "$dirprg/../corelib";
	unshift @INC, "$dirprg/../lib";
	$ENV{PATH} .= ":$dirprg";
	$VERSION = do {my @r = (q$Rev: 1613 $ =~ /\d+/g); sprintf "%d." . "%02d" x $#r, @r}
}

use strict;
use warnings;
use ParamParser;
use General;
use GeneralBioinfo;
use Carp;

my $MINLEN = 50;

MAIN:
{
    my $o_param = New ParamParser;
    $o_param->SetUsage('USAGE_DELAYED');
    $o_param->Update("GETOPTLONG", "I",( "infile=s","outfile=s","help","source=s","minlen=i","addparent","split=s"));
    $o_param->SetUsage(my $usage = sub {&Usage();});

    $o_param->AssertFileExists('infile');
	my $infile = $o_param->Get('infile');
    $o_param->SetUnlessDefined('outfile', $infile . ".N.gff3");
    my $outfile    = $o_param->Get('outfile');

    $o_param->SetUnlessDefined('minlen', "50");
	my $minlen = $o_param->Get('minlen');
    $o_param->SetUnlessDefined('source', "Nx$minlen");
	my $source = $o_param->Get('source');

    my $seqfile = $o_param->Get('infile');

	my @a_outfiles = ();
	my $split=&FALSE;
	my %h_outfiles=();
	my $f_gff3 = undef;
	if ( $o_param->IsDefined('split') )
	{
		$o_param->AssertFileExists('split');
		$split=&TRUE;
		my $f_split = $o_param->GetStreamIn('split');
		while(my $l=<$f_split>)
		{
			chomp($l);
			my ($dir,$file) = $l =~ /(.+)\/(.+)/;
			$h_outfiles{$file}=$l;
		}
		$f_split->close;
	}
	else
	{
		$f_gff3 = $o_param->GetStreamOut('outfile');
	}

	my %h_seq=();
	&FastaToHash($infile,\%h_seq);

	foreach my $seqname (sort keys(%h_seq))
	{
		my $seq = $h_seq{$seqname}{sequence};
		my $cpt=1;
		if ( defined($h_outfiles{$seqname}) )
		{
			$f_gff3 = &GetStreamOut("$h_outfiles{$seqname}.N.gff3");
		}
		while($seq =~ /(N{$minlen,})/ig )
		{
			my $pos = length($`)+1;
			my $len = length($1);
			print $f_gff3 "$seqname\t$source\tregion\t$pos\t",$pos+$len-1,"\t.\t.\t.\tID=$seqname:N$cpt;Name=$seqname:N$cpt\n"; 
			if ( $o_param->Get('addparent') )
			{
				print $f_gff3 "$seqname\t$source\tregion\t$pos\t",$pos+$len-1,"\t.\t.\t.\tID=C$seqname:N$cpt;Parent=$seqname:N$cpt;Name=C$seqname:N$cpt\n"; 
			}
			$cpt++;
		}
		if ( defined($h_outfiles{$seqname}) )
		{
			$f_gff3->close;
		}
	}
	$f_gff3->close if ( defined($f_gff3) );
}

=item Usage

	Print the usage of the program

=cut

sub Usage
{
    print STDERR<<END
$0     
Convert a fasta sequence files in gff3 file(s) containing regions of N streches (chado compliant)

Mandatory
--infile  filename   fasta sequence

Optionnal
--outfile     filename  gff3-formatted file
--split       filename  FileOfFiles of sequence files (will be used as prefix)
--minlen      integer   [$MINLEN]
--source      string    [Nx$MINLEN]
--addparent             flag to add parent features in gff3

WARNING: in split mode, fasta files in the fof must correspond to the sequence IDs (cf lipm_fasta2tree.pl)
END
      ;

}

