#!/usr/bin/perl

#
# $Id: lipm_tab_to_tab_nr.pl 1615 2018-07-19 06:09:48Z gouzy $
#

# Copyright INRA

# Jerome.Gouzy@toulouse.inra.fr
# Erika.Sallet@toulouse.inra.fr
# Sebastien.Carrere@toulouse.inra.fr
# Ludovic.Legrand@toulouse.inra.fr
# Ludovic.Cottret@toulouse.inra.fr

# This software aims to remove the redondancy in similaties analyses
# at the genome scale

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod 

=head1 NAME



=head1 SYNOPSIS



=head1 DESCRIPTION


# some files can be very big, use in priority optimized programs (perl -ne/sort)

=cut

BEGIN
{
	use File::Basename;
	my $dirprg=dirname($0);
	unshift @INC, "$dirprg/../corelib";
	$VERSION = do { my @r = (q$Rev: 1615 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r }
}

use strict;
use warnings;
use Carp;
use ParamParser;
use General;
use GeneralBioinfo;
use Data::Dumper;

my $DEF_QCOL         = 0;
my $DEF_FMT          = 'm8';
my $DEBUG            = 0;
my $DEF_PCS_OMAP     = 0.80;
my $DEF_MINLEN_OMAP  = 0;
my $DEF_PCS_M89      = 0.95;
my $DEF_MINLEN_M89   = 30; # for HSPs against protein

MAIN:
{
	 # Get parameters
    my $o_param = New ParamParser('GETOPTLONG', \&Usage,"in=s", "out=s",'switch','intra','multi-hits','fmt=s','pcs=s','debug','sort','minlen=i','nb-cols=i');

	$DEBUG = 1 if ( $o_param->IsDefined('debug') );

	$o_param->AssertDefined('in', 'out');
	$o_param->SetUnlessDefined('querycol',$DEF_QCOL);
	$o_param->SetUnlessDefined('fmt',$DEF_FMT);
	$o_param->AssertAllowedValue('fmt','m8','m9','omap');
	&GetExePath("intersectBed");

	my $qcol=$o_param->IsDefined('switch');
	$qcol++; 

	my $inputfile=$o_param->Get('in');
	my $fmt=$o_param->Get('fmt');
	my $multihits   = $o_param->IsDefined('multi-hits');

	my %h_multihits_ok=();
	my ($iqname,$iqstart,$iqend,$itname,$itscore,$bedfile) = &BuildDecreasingScoreSortedBed($o_param,\%h_multihits_ok);
	my $nb_col = $o_param->Get('nb-cols');

#	print STDERR Dumper(%h_multihits_ok);

	my %h_redundant=();
	&GetRedundantHits($o_param,$bedfile,\%h_redundant);

#	print STDERR Dumper(%h_redundant);

	my $fh_in    = $o_param->GetStreamIn('in');
	my $fh_out   = $o_param->GetStreamOut('out');
	if ( ! $o_param->IsDefined('sort') && $o_param->IsDefined('header') )
	{
		print $fh_out $o_param->Get('header')."\n";
	}
	my $minlen = $o_param->Get('minlen');
	my %h_hit;
	my %h_deja=();
	while( my $line = <$fh_in> )
	{
		if ( $line =~ /^(idmap|ref_id)/ )
		{
			#print $fh_out $line;
			next;
		}
        chomp ($line);
		my @a_ = split("\t",$line);
		next if ( scalar(@a_) != $nb_col );
		if ( $fmt =~ /m[89]/i )
		{
			&ParseM8Line(\$line,$qcol,\%h_hit);
		}
		else
		{
			$h_hit{query}  = $a_[$iqname];
			$h_hit{qend}   = $a_[$iqend];
			$h_hit{qstart} = $a_[$iqstart];
			$h_hit{score}  = $a_[$itscore];
			$h_hit{len}    = $h_hit{qend}-$h_hit{qstart}+1;
		}
		$h_hit{score}=sprintf("%d",$h_hit{score}*100);

		next if ( ! defined($h_hit{qend}) );
		my ($keymulti,$vmulti,$key)=();
		if ( $h_hit{qstart}<$h_hit{qend} )
		{
			$keymulti = "$h_hit{query}-$h_hit{qstart}";
			$vmulti   = sprintf("$h_hit{qstart}-$h_hit{qend}-%d",$h_hit{score});
			$key = "$h_hit{query} $h_hit{qstart} $h_hit{qend}"
		}
		else
		{
			$keymulti = "$h_hit{query}-$h_hit{qend}";
			$vmulti   = sprintf("$h_hit{qend}-$h_hit{qstart}-%d",$h_hit{score});
			$key     = "$h_hit{query} $h_hit{qend} $h_hit{qstart}";
		}
		if ($multihits)
		{
			next if ( ! exists($h_multihits_ok{$keymulti}) || $h_multihits_ok{$keymulti} ne $vmulti );
		} 

		next if ( $h_hit{len} < $minlen );


		if ( exists($h_redundant{$key}) )
		{
			if ($multihits)
			{
				next if ( $h_hit{score} < $h_redundant{$key} );
			}
			else
			{
				# 20180717
				# next;
				# as blastn can generate very large HSPs for the same pair of regions (same score in scientific notation but #mismastchs and #gaps differents)
				# it is safer to keep all of these hits if they have the best score
				next if ( $h_hit{score} < $h_redundant{$key} );
			}
		}

		next if ( exists($h_deja{$key}) );
		$h_deja{$key}=1;

		@a_ = split("\t",$line);
		if ($fmt =~ /m[89]/i )
		{
			$a_[11] = sprintf("%d",$a_[11]); # some blast scores can be in scientific notation, switch to decimal notation
		}
		print $fh_out join("\t",@a_)."\n";
	}
	$fh_in->close();
	$fh_out->close();
	if ( $o_param->IsDefined('sort') )
	{
		my $out = $o_param->Get('out');
		my $pos = $iqstart+1;
		my $nm  = $iqname+1;
		if ( $o_param->IsDefined('header') )
		{
			my $fh_out = &GetStreamOut("$out.sorted");
			print $fh_out $o_param->Get('header')."\n";
			$fh_out->close;
		}
		`grep -v ^idmap $out | grep -v ^ref_id | sort -k $pos,${pos}n | sort -s -k $nm,$nm >> $out.sorted`;
	}
}

#
# 
#

sub BuildDecreasingScoreSortedBed
{
my($o_param,$rh_multihits_ok)=@_;

	my $cde = "";
	my $fmt=$o_param->Get('fmt');
	my $nb_col=();
	my $inputfile=$o_param->Get('in');
	my $bedfile="$inputfile.$$.".time.".bed";
	my($iqname,$iqstart,$iqend,$itname,$itscore)=();
	my $switch = $o_param->IsDefined('switch');
	my $multihits = $o_param->IsDefined('multi-hits');
	my $fact = 1; # for hits vs optical maps the score is multiply by 100 to convert the float score into int
	if ($fmt =~ /m[89]/i )
	{
		($iqname,$iqstart,$iqend,$itname,$itscore)=($switch==0) ? (0,6,7,1,11) : (1,8,9,0,11);
		# for very long HSPs the score is reported in scientific notation, it must be reformated before sorting
		$cde = 'perl -ne \'chomp;@a_ = split(/\t/); $a_[11] = sprintf("%d",$a_[11]*100); print join("\t",@a_)."\n"; \' '.$inputfile.'  | sort -k12,12nr';
		$o_param->SetUnlessDefined('nb-cols',12);
		$o_param->SetUnlessDefined('pcs',$DEF_PCS_M89);
		$o_param->SetUnlessDefined('minlen',$DEF_MINLEN_M89);
		$fact = 1;  # the multiplication by 100 is performed here, no need to do hit again
	}
	else # omap
	{
#idmap   nb_sites        om_start        om_end  omlen   om_pos_start    om_pos_end      strand  idseq   seq_sites       seq_start       seq_stop        seqlen  seq_pos_start   seq_pos_end     ali_len score   density
#i2_ctg2955_r      154  22    38  1343686  216203  353240 - HanXRQCP                     19  2   18   147171  3093    143668  17   10.78   0.63    95
#OcIN23-M4-r1c005 2434 749  1121 16266687 4824172 7233371 + exp_refineFinal1_contig14_r 737  8  332  6383801 44595   2767388 373   208.91  0.5601  14.81   42.65

		# to be able to merge hits against optical maps (vs.omap), the same principle used in --multihits (developed for proteins) is used here. Only the best-scoring hit of a region is kept
		($iqname,$iqstart,$iqend,$itname,$itscore)=($switch==0) ? (0,5,6,8,16) : (8,13,14,0,16);
		$cde = "grep -v ^idmap $inputfile | grep -v ^ref_id | sort -k 17,17nr ";
		$o_param->SetUnlessDefined('nb-cols',19);
		$o_param->SetUnlessDefined('pcs',$DEF_PCS_OMAP);
		$o_param->SetUnlessDefined('minlen',$DEF_MINLEN_OMAP);
		my @a_header = `grep ^idmap $inputfile | grep -v ^ref_id | head -n 1`;
		chomp(@a_header);
		$o_param->SetUnlessDefined('header',$a_header[0]) if ( scalar(@a_header) );
		$fact = 100;
	}
	if ( $multihits )
	{
		# as HSPs are sorted by decreasing score, only the first hit is kept for a start position
		$cde .= ' | perl -ane \'$k=($F['.$iqstart.']<$F['.$iqend.'])?"$F['.$iqname.']-$F['.$iqstart.']":"$F['.$iqname.']-$F['.$iqend.']";print if ( ! $h_deja{$k} );$h_deja{$k}=1;\'';
	}
	# here files are 1-based
	$cde .= ' | perl -ane \'next if ( ! defined($F['.$itscore.']) || /^(idmap|ref_id)/ ); $sc = sprintf("%d",$F['.$itscore.']*'.$fact.'); if ( $F['.$iqstart.'] < $F['.$iqend.'] ) { print "$F['.$iqname.']\t$F['.$iqstart.']\t$F['.$iqend.']\t$F['.$itname.']___$sc\n" } else { print "$F['.$iqname.']\t$F['.$iqend.']\t$F['.$iqstart.']\t$F['.$itname.']___$sc\n" } \' > '.$bedfile.'';
	&System($cde);
	#  at that stage, the score was multiplied by 100 for both the blast m8 format and hits against optical maps

	# for a genomic position, only was kept in the bed the best scoring hit. These HSPs are stored in the hash
	if ( $multihits )
	{
		my $f_in = &GetStreamIn("$bedfile");
		while(my $l=<$f_in>)
		{
			chomp($l);
			my @a_=split(' ',$l);
			my $k = "$a_[0]-$a_[1]";
			my ($score) = $a_[3] =~ /___(\d+)$/;
			# here the score was *100 (to convert the float score vs omap into int)
			$$rh_multihits_ok{$k}="$a_[1]-$a_[2]-$score";
		}
		$f_in->close;
	}

	return($iqname,$iqstart,$iqend,$itname,$itscore,$bedfile);
}


#
#
#

sub GetRedundantHits
{
my($o_param,$bedin,$rh_redundant)=@_;

	my $intra     = $o_param->IsDefined('intra');
	my $multihits = $o_param->IsDefined('multi-hits');
	my $pcs       = $o_param->Get('pcs');

	my $bedout = "$bedin.redundant.bed";
	if ( ! $intra )
	{
			#if ( ! $multihits )
			#
			# compute how many hits are included in each region (at the -f cutoff)
			# the fast option below, purely based on bedtools, postulates that the actual corresponding region of the target will always be the best scoring 
			# hit that can overlap at the --pcs/-f cutoff (0.95) on the query only with itself (count = 1) and that regions that overlap several times (including itself) 
			# are redundant regions
	 	    # &System("intersectBed -a $bedin -b $bedin -f $pcs -c | perl -ane 'print if ( \$F[4] > 1 )' > $bedout");
			# but the assumption can be wrong with blastn that sometimes generates very large overlapping HSPs with a maximum score
			#
			#else
			#
			# store the number of hits corresponding to included regions
			# included regions are associated to the score on the including best-score
			# it is much slower than the previous version but can handle several hits with a maximum score 
			#&System("intersectBed -a $bedin -b $bedin -f $pcs -wo > $bedout.raw"); # the output file can use tens of Gb
			my $f_in  = &GetStreamIn("intersectBed -a $bedin -b $bedin -f $pcs -wo|");
			my %h_reg =();
			my %h_deja =();
			while(my $l=<$f_in>)
			{
				chomp($l);
				my @a_ = split(/\t/,$l);
				my($currregscore,$tgtregscore) = $l =~ /___(\S+).+___(\S+)/;
				my($max)=&Max($currregscore,$tgtregscore);
				my $key = "$a_[0]\t$a_[1]\t$a_[2]";
				my $key2= "$a_[4]\t$a_[5]\t$a_[6]";
				next if ( $key eq $key2 );
				# max score propagation on redundant regions (at the end, the curr score will be compared to the maximum score that is associated with and removed if lower)
				if (  ! defined($h_reg{$key}) || $h_reg{$key} < $max )
				{
					$h_reg{$key} = $max;
				}
			}
			my $f_out = &GetStreamOut("$bedout");
			foreach my $key (keys(%h_reg))
			{
				print $f_out "$key\t$h_reg{$key}\n";
			}
			$f_out->close;
			$f_in->close;
			#unlink("$bedout.raw") if ( ! $DEBUG );
			#
	}
	else # removing redundancy in hits of a reference sequence, can be stressfull for the filesystem
	{
		&System("rm -f $bedout");
		my $c=0;
		my %h_deja = ();
		my $f_bed  = &GetStreamIn($bedin);
		while(my $f=<$f_bed>) # first loop on the query
		{
			chomp($f);
			my ($seqid,@a_tmp) = split(' ',$f);

			next if ( exists($h_deja{$seqid}) );
			$h_deja{$seqid}=1;

			$c++;
			print STDERR "$seqid\t$c\n" if ( $DEBUG );

			# build a bed file with only hits of seqid (requires differents Ids in query and target)
			my $outtmp1="$bedin.$seqid.bed";
			&System("grep -w $seqid $bedin > $outtmp1");
			my %h_count=();
			my $cpt=0;

			my $f_bed2 = &GetStreamIn($outtmp1);
			while(my $l2=<$f_bed2>)	# second loop on the target
			{
				my ($t0,$t1,$t2,$seqid2) = split(' ',$l2);
				$h_count{$seqid2}++;
				$cpt++;
			}
			$f_bed2->close;

			if ( $cpt < 2 ) # not redundant, the hit is not included in another one
			{
				&System("rm -f $outtmp1");
				next;
			}

			my %h_deja_target = ();
			$f_bed2 = &GetStreamIn($outtmp1);
			while(my $l2=<$f_bed2>)	# second loop on the target
			{
				my ($t0,$t1,$t2,$seqid2) = split(' ',$l2);
				next if ( exists($h_deja_target{$seqid2}) );
				next if ($h_count{$seqid2} == 1);
				$h_deja_target{$seqid2}=1;
				my $outtmp12="$bedin.$seqid.$seqid2.bed";
				&System("grep -w $seqid2 $outtmp1 > $outtmp12");
				&System("intersectBed -a $outtmp12 -b $outtmp12 -f $pcs -c | perl -ane 'print if ( \$F[4] > 1 )' >> $bedout");
				&System("rm -f $outtmp12");
			}
			$f_bed2->close;
			&System("rm -f $outtmp1");
		}
		$f_bed->close;
	}
	# Load redundant hits
	my $f_in = &GetStreamIn("$bedout");
	while(my $l=<$f_in>)
	{
		chomp($l);
		my @a_=split(' ',$l);
		my $k = "$a_[0] $a_[1] $a_[2]";
		$$rh_redundant{$k}= $a_[3] || 1;
	}
	$f_in->close;
	unlink($bedout) if ( ! $DEBUG );
	unlink($bedin)  if ( ! $DEBUG ) ;
}

#
#
#

sub System
{
my($cde)=@_;

	print STDERR "$cde\n" if ($DEBUG);
	system($cde);
}

=item Usage

	$o_param->Usage();
	Print the usage of the program

=cut

sub Usage
{
    print STDERR<<END

Remove hits included in larger/stronger/better hits on the same genomic region (based on intersectBed)

Reciprocal best hits are retrieved in two passes
$0 --in file     --out file.nr1              [--fmt omap]
$0 --in file.nr1 --out file.nr1.nr2 --switch [--fmt omap]

Scope: the default parameters assume that the query (first column) is a genome

Usage:
$0	
  --help                   this message
[Mandatory]
  --in      filename       tabulated file, m8|omap formatted
  --out     filename       tabulated file, m8|omap formatted
[Optional]
  --fmt      string        [$DEF_FMT] (m8|omap)
  --intra                  remove redundancy between a given pair of query/target (useful to remove secondary hits, require different Ids in query and target)
  --multi-hits             use this flag to manage an expected redundancy in the target (e.g: hits with a protein database|mapping of several genome versions vs same omap)
  --pcs      float         cf intersectBed -f ]0,1] default are $DEF_PCS_OMAP/omap and $DEF_PCS_M89/m8
  --switch                 flag to switch the reference
END
}

